
/* Chrysalide - Outil d'analyse de fichiers binaires
 * trigger.c - équivalent Python du fichier "plugins/kaitai/rost/trigger.c"
 *
 * Copyright (C) 2023 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "trigger.h"


#include <assert.h>
#include <pygobject.h>


#include <i18n.h>
#include <plugins/pychrysalide/access.h>
#include <plugins/pychrysalide/helpers.h>
#include <plugins/pychrysalide/analysis/scan/item.h>


#include "../parsers/struct.h"
#include "../../parsers/struct.h"
#include "../../rost/trigger-int.h"



CREATE_DYN_CONSTRUCTOR(kaitai_trigger, G_TYPE_KAITAI_TRIGGER);

/* Initialise une instance sur la base du dérivé de GObject. */
static int py_kaitai_trigger_init(PyObject *, PyObject *, PyObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : self = objet à initialiser (théoriquement).                  *
*                args = arguments fournis à l'appel.                          *
*                kwds = arguments de type key=val fournis.                    *
*                                                                             *
*  Description : Initialise une instance sur la base du dérivé de GObject.    *
*                                                                             *
*  Retour      : 0.                                                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int py_kaitai_trigger_init(PyObject *self, PyObject *args, PyObject *kwds)
{
    GKaitaiStruct *kstruct;                 /* Définition Kaitai en place  */
    int ret;                                /* Bilan de lecture des args.  */
    GKaitaiTrigger *trigger;                /* Création GLib à transmettre */

#define KAITAI_TRIGGER_DOC                                                      \
    "The KaitaiTrigger object store an access to a loaded Kaitai definition"    \
    " creates as many as required instances from it when ROST filters contents" \
    " with the help of the Kaitai module.\n"                                    \
    "\n"                                                                        \
    "Instances can be created using following constructor:\n"                   \
    "\n"                                                                        \
    "    KaitaiTrigger(kstruct)"                                                \
    "\n"                                                                        \
    "Where *kstruct* is a pychrysalide.plugins.kaitai.parsers.KaitaiStruct"     \
    " instance loaded with a valid Kaitai definition."

    /* Récupération des paramètres */

    ret = PyArg_ParseTuple(args, "O&", convert_to_kaitai_structure, &kstruct);
    if (!ret) return -1;

    /* Initialisation d'un objet GLib */

    ret = forward_pygobjet_init(self);
    if (ret == -1) return -1;

    /* Eléments de base */

    trigger = G_KAITAI_TRIGGER(pygobject_get(self));

    if (!g_kaitai_trigger_create(trigger, kstruct))
    {
        PyErr_SetString(PyExc_ValueError, _("Unable to create Kaitai stream."));
        return -1;

    }

    return 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_kaitai_trigger_type(void)
{
    static PyMethodDef py_kaitai_trigger_methods[] = {
        { NULL }
    };

    static PyGetSetDef py_kaitai_trigger_getseters[] = {
        { NULL }
    };

    static PyTypeObject py_kaitai_trigger_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.plugins.kaitai.rost.KaitaiTrigger",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,

        .tp_doc         = KAITAI_TRIGGER_DOC,

        .tp_methods     = py_kaitai_trigger_methods,
        .tp_getset      = py_kaitai_trigger_getseters,

        .tp_init        = py_kaitai_trigger_init,
        .tp_new         = py_kaitai_trigger_new,

    };

    return &py_kaitai_trigger_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide....rost.KaitaiTrigger. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool ensure_python_kaitai_trigger_is_registered(void)
{
    PyTypeObject *type;                     /* Type Python 'RecordEmpty'   */
    PyObject *module;                       /* Module à recompléter        */
    PyObject *dict;                         /* Dictionnaire du module      */

    type = get_python_kaitai_trigger_type();

    if (!PyType_HasFeature(type, Py_TPFLAGS_READY))
    {
        module = get_access_to_python_module("pychrysalide.plugins.kaitai.rost");

        dict = PyModule_GetDict(module);

        if (!ensure_python_scan_registered_item_is_registered())
            return false;

        if (!register_class_for_pygobject(dict, G_TYPE_KAITAI_TRIGGER, type))
            return false;

    }

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : arg = argument quelconque à tenter de convertir.             *
*                dst = destination des valeurs récupérées en cas de succès.   *
*                                                                             *
*  Description : Tente de convertir en accès à une définition Kaitai.         *
*                                                                             *
*  Retour      : Bilan de l'opération, voire indications supplémentaires.     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

int convert_to_kaitai_trigger(PyObject *arg, void *dst)
{
    int result;                             /* Bilan à retourner           */

    result = PyObject_IsInstance(arg, (PyObject *)get_python_kaitai_trigger_type());

    switch (result)
    {
        case -1:
            /* L'exception est déjà fixée par Python */
            result = 0;
            break;

        case 0:
            PyErr_SetString(PyExc_TypeError, "unable to convert the provided argument to Kaitai trigger");
            break;

        case 1:
            *((GKaitaiTrigger **)dst) = G_KAITAI_TRIGGER(pygobject_get(arg));
            break;

        default:
            assert(false);
            break;

    }

    return result;

}
